/*
MPEG 2 decoder bitstream xfer
*/

#include <stdio.h>
#include <stdlib.h>

#define	TRUE		1
#define	FALSE		0

void conv444to422(unsigned char *src, unsigned char *dst, int width, int height);

unsigned char *Clip;

int main(int argc, char **argv)
{
	FILE	*fpout;
	int	length, i, j, horz, vert, index = 0;
	int	r, g, b;
	int	red, green, blue;
	double	y, u, v;
	double	cr, cg, cb, cu, cv;
	unsigned char	*yp, *up, *vp;
	int	matrix_coefficients = 1;
	unsigned int	py, pcb, pcr;
	static unsigned char	buffer_422output[1920 * 1088 * 2];
	static unsigned char	buffer_422Cb[1920 * 1088 / 2];
	static unsigned char	buffer_422Cr[1920 * 1088 / 2];
	static unsigned char	buffer_444Y[1920 * 1088];
	static unsigned char	buffer_444Cb[1920 * 1088];
	static unsigned char	buffer_444Cr[1920 * 1088];
	static double coef[7][3] = {
		{0.2126,0.7152,0.0722}, /* ITU-R Rec. 709 (1990) */
		{0.299, 0.587, 0.114},  /* unspecified */
		{0.299, 0.587, 0.114},  /* reserved */
		{0.30,  0.59,  0.11},   /* FCC */
		{0.299, 0.587, 0.114},  /* ITU-R Rec. 624-4 System B, G */
		{0.299, 0.587, 0.114},  /* SMPTE 170M */
		{0.212, 0.701, 0.087}}; /* SMPTE 240M (1987) */

	if (argc != 4) {
		fprintf(stderr, "usage: store <horz> <vert> <outfile>\n");
		exit(-1);
	}

	/*--- open binary file (for parsing) ---*/
	fpout = fopen(argv[3], "w");
	if (fpout == 0) {
		fprintf(stderr, "Cannot open output file <%s>\n", argv[4]);
		exit(-1);
	}

	/* Clip table */
	if (!(Clip=(unsigned char *)malloc(1024)))
		printf("Clip[] malloc failed\n");

	Clip += 384;

	for (i = -384; i < 640; i++)
		Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

	horz = atoi(argv[1]);
	vert = atoi(argv[2]);

	i = matrix_coefficients;
	cr = coef[i-1][0];
	cg = coef[i-1][1];
	cb = coef[i-1][2];
	cu = 0.5/(1.0-cb);
	cv = 0.5/(1.0-cr);

	for (i = 0; i < vert; i++)
	{
		blue = 0;
		green = 0;
		red = 0;
		yp = &buffer_444Y[0] + i*horz;
		up = &buffer_444Cb[0] + i*horz;
		vp = &buffer_444Cr[0] + i*horz;

		for (j = 0; j < horz; j++)
		{
			if (i < 296)  {
				if ((j >= 64) && (j < 1856))  {
					red = (j - 64) / 7;
					blue = 0;
					green = 0;
				}
				else if (j == 62 || j == 63 || j == 1856 || j == 1857)  {
					blue = 255;
					green = 255;
					red = 255;
				}
				else  {
					blue = 0;
					green = 0;
					red = 0;
				}
			}
#if 0
			else if ((i >= 352) && (i < 364))  {
				if ((j >= 64) && (j < 1856))  {
					if (((j - 64) % 7) == 0)  {
						blue = 255;
						green = 255;
						red = 255;
					}
					else if (((j - 65) % 7) == 0)  {
						blue = 255;
						green = 255;
						red = 255;
					}
					else  {
						blue = 0;
						green = 0;
						red = 0;
					}
				}
				else  {
					blue = 0;
					green = 0;
					red = 0;
				}
			}
#else
			else if ((i >= 296) && (i < 392))  {
				if ((j >= 64) && (j < 1856))  {
					red = (j - 64) / 7;
					green = (j - 64) / 7;
					blue = (j - 64) / 7;
				}
				else if (j == 62 || j == 63 || j == 1856 || j == 1857)  {
					blue = 255;
					green = 255;
					red = 255;
				}
				else  {
					red = 0;
					green = 0;
					blue = 0;
				}
			}
#endif
			else if ((i >= 392) && (i < 688))  {
				if ((j >= 64) && (j < 1856))  {
					green = (j - 64) / 7;
					red = 0;
					blue = 0;
				}
				else if (j == 62 || j == 63 || j == 1856 || j == 1857)  {
					blue = 255;
					green = 255;
					red = 255;
				}
				else  {
					red = 0;
					green = 0;
					blue = 0;
				}
			}
#if 0
			else if ((i >= 716) && (i < 728))  {
				if ((j >= 64) && (j < 1856))  {
					if (((j - 64) % 7) == 0)  {
						blue = 255;
						green = 255;
						red = 255;
					}
					else if (((j - 65) % 7) == 0)  {
						blue = 255;
						green = 255;
						red = 255;
					}
					else  {
						blue = 0;
						green = 0;
						red = 0;
					}
				}
				else  {
					blue = 0;
					green = 0;
					red = 0;
				}
			}
#else
			else if ((i >= 688) && (i < 784))  {
				if ((j >= 64) && (j < 1856))  {
					red = (j - 64) / 7;
					green = (j - 64) / 7;
					blue = (j - 64) / 7;
				}
				else if (j == 62 || j == 63 || j == 1856 || j == 1857)  {
					blue = 255;
					green = 255;
					red = 255;
				}
				else  {
					red = 0;
					green = 0;
					blue = 0;
				}
			}
#endif
			else if (i >= 784) {
				if ((j >= 64) && (j < 1856))  {
					blue = (j - 64) / 7;
					red = 0;
					green = 0;
				}
				else if (j == 62 || j == 63 || j == 1856 || j == 1857)  {
					blue = 255;
					green = 255;
					red = 255;
				}
				else  {
					red = 0;
					green = 0;
					blue = 0;
				}
			}
			b = blue;
			g = green;
			r = red;
			/* convert to YUV */
			y = cr * r + cg * g + cb * b;
			u = cu * (b - y);
			v = cv * (r - y);
			yp[j] = (219.0 / 256.0) * y + 16.5;  /* nominal range: 16..235 */
			up[j] = (224.0 / 256.0) * u + 128.5; /* 16..240 */
			vp[j] = (224.0 / 256.0) * v + 128.5; /* 16..240 */
#if 0
			printf("h = %d, r = %d, g = %d, b = %d, y = %d, Cb = %d, Cr = %d\n", j,r,g,b,yp[j],up[j],vp[j]);
#endif
		}
	}
	conv444to422(&buffer_444Cb[0], &buffer_422Cb[0], horz, vert);
	conv444to422(&buffer_444Cr[0], &buffer_422Cr[0], horz, vert);
	py = 0;
	pcb = 0;
	pcr = 0;
	for (i = 0; i < (horz * vert * 2); i++)
	{
		if ((i % 2) == 1)
		{
			buffer_422output[i] = buffer_444Y[py++];
		}
		else if ((i % 4) == 0)
		{
			buffer_422output[i] = buffer_422Cb[pcb++];
		}
		else
		{
			buffer_422output[i] = buffer_422Cr[pcr++];
		}
	}
	fwrite(&buffer_422output[0], 1, (horz * vert * 2), fpout);
	fclose(fpout);
	return 0;
}

/* horizontal filter and 2:1 subsampling */
void conv444to422(unsigned char *src, unsigned char *dst, int width, int height)
{
  int i, j, im5, im4, im3, im2, im1, ip1, ip2, ip3, ip4, ip5, ip6;

  if (0)
  {
    for (j=0; j<height; j++)
    {
      for (i=0; i<width; i+=2)
      {
        im5 = (i<5) ? 0 : i-5;
        im4 = (i<4) ? 0 : i-4;
        im3 = (i<3) ? 0 : i-3;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<width-1) ? i+1 : width-1;
        ip2 = (i<width-2) ? i+2 : width-1;
        ip3 = (i<width-3) ? i+3 : width-1;
        ip4 = (i<width-4) ? i+4 : width-1;
        ip5 = (i<width-5) ? i+5 : width-1;
        ip6 = (i<width-5) ? i+6 : width-1;

        /* FIR filter with 0.5 sample interval phase shift */
        dst[i>>1] = Clip[(int)(228*(src[i]+src[ip1])
                         +70*(src[im1]+src[ip2])
                         -37*(src[im2]+src[ip3])
                         -21*(src[im3]+src[ip4])
                         +11*(src[im4]+src[ip5])
                         + 5*(src[im5]+src[ip6])+256)>>9];
      }
      src+= width;
      dst+= width>>1;
    }
  }
  else
  {
    /* MPEG-2 */
    for (j=0; j<height; j++)
    {
      for (i=0; i<width; i+=2)
      {
        im5 = (i<5) ? 0 : i-5;
        im3 = (i<3) ? 0 : i-3;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<width-1) ? i+1 : width-1;
        ip3 = (i<width-3) ? i+3 : width-1;
        ip5 = (i<width-5) ? i+5 : width-1;

        /* Original FIR filter coefficients (*512): 22 0 -52 0 159 256 159 0 -52 0 22 */
        /* It's a bug. Filter coefficients must add up to 512!!! */

        /* FIR filter coefficients (*512): 21 0 -52 0 159 256 159 0 -52 0 21 */
        dst[i>>1] = Clip[(int)(  21*(src[im5]+src[ip5])-52*(src[im3]+src[ip3])
                         +159*(src[im1]+src[ip1])+256*src[i]+256)>>9];
      }
      src+= width;
      dst+= width>>1;
    }
  }
}
